#include "stdio.h"
#include "stdlib.h"
#include "task.h"
#include "enable.h"
#include "main.h"
#include "pq.h"
#include "malloc_m.h"
#include "BlockAlloc.h"
extern volatile int Tick;
int TStamp;
volatile int InterruptCount;
volatile int Blocking;
TCB *pxCurrentTCB,*pxReadyTCB;
int OS_IntNesting;
PQ ActiveTasks;
ECB *TimeoutList;
static ECB *ECBPool;

/***********************************************************************
** TimerTicker                                                        **
**                                                                    **
**	This function is called by the real time interrupt handler        **
**	This function Increments the clock, scans through the Event       **
**  Control Blocks looking for timed out events.                      **
**                                                                    **
***********************************************************************/


void TimerTicker(void)
{
	int sr;
	++TStamp;
	//we need to process the list of tasks waiting
	//for queues
	ECB *e;
	TCB *t,**prev;

	sr = Disable();
	for(e = TimeoutList;e;e=e->Tnext)	//go through whole list
	{
		for(prev = &(e->task_h),t = e->task_h;t;)
		{
			if(t->timeout > 0)	//if less than zero, infinate timeout
			{
				if( --(t->timeout) <= 0)	//decrement timeout counter
				{
					t->status = EVENT_TIMEOUT;	//event caused by semaphore
					*prev = t->next;
					t->next = 0;
					Insert(&ActiveTasks,t);
				}
			}
			prev = &(t->next);
			t = t->next;
		}
	}
	Enable(sr);
}

//----------------------------------------------------------------------
//
// This function is used to compare two items in the priority queue
// to see which one has the highest priority
//
// The function is passes two pointers that point to the pointers
// that point to the actual objects
//
//----------------------------------------------------------------------
//
static int PriorityCompare(void **s1, void **s2)
{
	int r;

	//check priority levels
	if( (r = ((TCB *)*s1)->priority - ((TCB *)*s2)->priority) != 0)
		return r;
	//if priority levels are the same, then check time stamps
	return (int)(((TCB *)*s2)->TimeStamp - ((TCB *)*s1)->TimeStamp);
}
/*
static int TaskCompare(void **s1,void **s2)
{
	int retval = 1;

	if(*s1 == *s2) retval = 0;
	return retval;
}
*/
void CreateTask(TCB *t, void (*task)(void *),long *stack, int stacksize,int priority,char *name,void *arg)
{
	/***************************************************************************************
	** Create a task
	** parameter:
	**	t...............pointer to the task control block
	**	task............pointer to function that will be the task
	**	stack...........pointer to memory block to use as stack
	**	stacksize.......size of stack in number of longs
	**	priority........task priority
	**	name............tag to identify task control block for debug purposes
	**	arg.............argument to be passed to function...it chan either be 
	**  a value, or a pointer to some data
	***************************************************************************************/
	long *stk;
	long sr;

	sr = Disable();
	if(t)
	{
		t->TcbSwaps = 0;
		t->stacksize = stacksize;
		t->priority = priority;
		t->TimeStamp = 0;
		t->name = name;
		t->stacktop = stack;
		t->next = (TCB *)0;
		t->list = (TCB *)0;
		stk = stack + stacksize;
		*--stk = (long )(task);/*	return address	*/
		*--stk = (long)0x14141414;		/*	r14 = 0	*/
		*--stk = 0x12121212;		/*	r12 = 0	*/
		*--stk = 0x11111111;		/*	r11 = 0	*/
		*--stk = 0x10101010;		/*	r10 = 0	*/
		*--stk = 0x09090909;		/*	r9 = 0	*/
		*--stk = 0x08080808;		/*	r8 = 0	*/
		*--stk = 0x07070707;		/*	r7 = 0	*/
		*--stk = 0x06060606;		/*	r6 = 0	*/
		*--stk = 0x05050505;		/*	r5 = 0	*/
		*--stk = 0x04040404;		/*	r4 = 0	*/
		*--stk = 0x03030303;		/*	r3 = 0	*/
		*--stk = 0x02020202;		/*	r2 = 0	*/
		*--stk = 0x01010101;		/*	r1 = 0	*/
		*--stk = (long)arg;				/*	r0 = pointer to passed arguments	*/
		*--stk = 0x1f; 				/* status reg:System mode, ARM mode, interrupts enabled. */
		t->stack = stk;
	}
	Enable(sr);
}

/************************************************************************
**                                                                     **
** ExitInterrupt                                                       **
**                                                                     **
** This function keeps track of the nesting level of interrupts        **
** If nesting level is == 0, then do a context swap to the highest     **
** priority task                                                       **
**                                                                     **
** WARNING!                                                            **
** 	Don't mess with this function, you will be sorry                   **
**  Adding any local variables or passed parameters                    **
**  Messes with the stack.  IrqSwap alters the stack and               **
**  if it is different from what it expects, it will mess              **
**  up the stack.                                                      **
**                                                                     **
** Hey, you don't even need to add any local variables, just add a     **
** Little bit more code and it will mess things up.  Always check      **
** the listing file to see how much might get pushed onto the stack	   **
**                                                                     **
**                                                                     **
** ExitInterrupt will alocate at least 4 bytes on the stack probably   **
** for sr, depending on how it does this.                              **
** Currently ExitInterrupt uses 4 bytes on the stack to save various   **
** registers that it doesn't really have to                            **
**                                                                     **
** Currently IrqSwap() needs to remove 12 bytes from stack			   **
**				                                                       **
************************************************************************/

void ExitInterrupt(void)
{
	int sr;

	--InterruptCount;
	sr = Disable();
	if(!InterruptCount && !Blocking)	//don't swap if nested
	{
		pxCurrentTCB->TimeStamp = TStamp;
		Replace(&ActiveTasks,(void **)&pxReadyTCB,pxCurrentTCB);
		if(pxReadyTCB != pxCurrentTCB)
		{
			pxReadyTCB->TcbSwaps++;
			IrqSwap();	//interrupts will be re-enabled right after this
		}
	}
	Enable(sr);
}

/*********************************************************************
** This function is used to yield (please note that the function name
** is misspelled on purpose, there is a Yield funciton somewhere in the
** IAR library) control of the processor to another task, but only if
** we are at the task level.  We cannot yield if we are in an interrupt
** routine
*********************************************************************/

void Yeild(void)
{
	long sr;

	if(!InterruptCount)
	{
		sr = Disable();
		pxCurrentTCB->TimeStamp = TStamp;
		Replace(&ActiveTasks,(void **)&pxReadyTCB,(void *)pxCurrentTCB);
		pxReadyTCB->TcbSwaps++;
		vPortYieldProcessor();
		Enable(sr);
	}
}

static ECB *AllocECB(void)
{
	//--------------------------------------------------------------------------
	// Allocate an event control Block
	//	Check the ECB pool, if it is empty, then allocate a new ECB from heap
	//
	// return value:
	//	pointer to ECB on success
	//	NULL on fail
	//--------------------------------------------------------------------------
	ECB *rv;
	
	int sr = Disable();
	if (ECBPool)	//check the pool
	{
		rv = ECBPool;
		ECBPool = ECBPool->Tnext;
		Enable(sr);
	}
	else
	{
		Enable(sr);
		rv = (ECB *)Malloc(sizeof(ECB));	//or allocate from the heap
	}
	return rv;
}

static void FreeECB(ECB *e)
{
	//--------------------------------------------------------------------------
	// put ECB block into ECB pool (don't bother to return it to the heap
	//
	// parameter:
	//	e...........pointer to ECB block to free
	//--------------------------------------------------------------------------
	int sr;

	sr = Disable();
	if(e->SemaphoreMode)	//is it in timeout list?
	{
		if(e == TimeoutList)
			TimeoutList = e->Tnext;
		if(e->Tprev)
			e->Tprev->Tnext = e->Tnext;
		if(e->Tnext)
			e->Tnext->Tprev = e->Tprev;
	}
	e->Tnext = ECBPool;
	ECBPool = e;
	Enable(sr);
}

ECB *NewSemaphore(int InitCount,int Mode,char *name)
{
	ECB *r;

	if((r = AllocECB()) != NULL)
		CreateSemaphore(r,InitCount,Mode,name);
	return r;
}

/***************************************************************************
**                                                                        **
** Semaphore objects
****************************************************************************
** CreateSemaphore                                                        **
** parameters:                                                            **
**	e............pointer to an event control block to initialize          **
**	InitCount....Initial value of semaphore count                         **
**	Mode.........0 = Semaphore doesn't time out, 1 = Semaphore Times out  **
**	n............name of semaphore (DEBUG only)                           **
**                                                                        **
***************************************************************************/

void CreateSemaphore(ECB *e,int InitCount,int Mode,char *n)
{
	int sr;

	e->EventCount = InitCount;
	e->SemaphoreMode = Mode;
	strcpy(e->name,n);					//copy name in
	e->task_h = 0;						//no tasks waiting
	e->task_t = 0;						//no tasks waiting
	//-------------------------------------------------------
	// if Mode==1, then put this semaphore into timeout list
	//-------------------------------------------------------
	sr = Disable();
	if(e->SemaphoreMode == 1)
	{
		if(TimeoutList)	//is list already started?
		{
			e->Tnext = TimeoutList;	//make current head next in line
			e->Tprev = 0;			//these is no previous entry
			TimeoutList->Tprev = e;	//this will be prev for next entry
			TimeoutList = e;			//this is now the new head of list
		}
		else			//no start list
		{
			TimeoutList = e;
			e->Tnext = 0;
			e->Tprev = 0;
		}
	}
	Enable(sr);	//restore interrupts
}

/*******************************************************************************
** Delete Semaphore
**	This function is used to free a Semaphore.  Rather than return the memory
** to the heap, it is put into a pool where it can be reused later without
** having to do another malloc.
**
**	parameters:
**		e..........pointer to ECB block that describes the semaphore
**
*******************************************************************************/

void DeleteSemaphore(ECB *e)
{
	int sr;
	TCB *t;

	//are there any tasks waiting for this sucker?
	sr = Disable();
	while(e->task_h)		//remove all tasks
	{
		t = e->task_h;		//get first task in
		e->task_h = t->next;       //remove from list
		t->next = (TCB *)0;
		t->status = EVENT_DELETED;			//return this value
		Insert(&ActiveTasks,t);	//put task back onto active list
	}
	FreeECB(e);		//put semaphore back in pool
	Enable(sr);
}

/**************************************************************************
** PendSemaphore (Lock)
**		This function checks the value of the semaphore.
**		1. Check Value EventCount
**		2. Decrement the value of the EventCount
**		3. If Value in Step 1 was greater than 0, then return
**		4. If Value in Step 1 was less than or equal to 0 THEN
**		5. Add task to list of pending tasks
**      6. Do a Context Switch
**
** parameters:
**	e.............pointer to event control block for semaphore
**	Timeout.......Number of system timer ticks to wait for semaphore to post
**
** return value:
**	Inidicates status of operation.  0 or positive is normal,
**		negative is error
**************************************************************************/

int PendSemaphore(ECB *e, int Timeout)	//wait for semaphore availiable
{
	int sr;
	int retval;
	sr = Disable();
	if(e->EventCount-- > 0)	/*	is resource availiable?	*/
	{
		/*	yes	*/
		Enable(sr);
		retval = EVENT_NOERROR;	//everything is peachy
	}
	else
	{
		/*	no	*/
		if(!e->task_h)		//make linked list of waiting tasks
			e->task_h = pxCurrentTCB;
		else
			e->task_t->next = pxCurrentTCB;
		e->task_t = pxCurrentTCB;
		pxCurrentTCB->status = EVENT_NOERROR;
		if(e->SemaphoreMode)
			pxCurrentTCB->timeout = Timeout;

		//swap out current task

		pxCurrentTCB->TimeStamp = TStamp;
		if(Delete(&ActiveTasks,(void **)&pxReadyTCB))	//get highest priority task
		{
			pxReadyTCB->TcbSwaps++;
			vPortYieldProcessor();                    //do context swap
		}
		else
		{
			pxCurrentTCB->status = EVENT_NOTASKS;
		}
		//-----------------------
		// Pend is over, check status
		//-----------------------

		Enable(sr);
		retval = pxCurrentTCB->status;	//return this value
	}
	return retval;
}

/***************************************************************************
** PostSemaphore (Unlock)
**
**		This function is used to incrment the value of a sempaphore
**		1. Check value of event flag
**		2. Increment event flag
**		3. If value is greater than or equal to 0, return
**		4. Else remove task in pending list and put it into the active task queue
**		5. return
**
** parameters:
**	e.................pointer to event control block for semaphore
**	Value.............Value to return to pending task (should be positive)
**
** return value: 0 = No Error, negative indicates error
***************************************************************************/

int PostSemaphore(ECB *e, int Value)	//signal semaphore availiable
{
	TCB *t;
	int sr;

	sr = Disable();
	if(e->EventCount < 32766)	//is it going to overflow??
	{
		if(e->EventCount++ >= 0)	//increment semaphore
		{
			Enable(sr);
		}
		else
		{
			if(e->task_h)				//is there really a task waiting?
			{
				t = e->task_h;		//get first task in
				e->task_h = t->next;       //remove from list
				t->next = (TCB *)0;
				t->status = Value;			//return this value
				Insert(&ActiveTasks,t);	//put task back onto active list
				Enable(sr);
				Yeild();
			}
			else
			{
				Enable(sr);
			}
		}
		return EVENT_NOERROR;
	}
	else
	{
		Enable(sr);
		return(EVENT_OVERFLOW);
	}
}

int TimeDelay(int mSec)
{
	//--------------------------------------------------------------------------
	// Delay task for a time
	//	parameters:
	//	mSec.............number of milliseconds to delay for
	//
	// return value:
	//	returns non zero value if something happened out of the ordinary
	//--------------------------------------------------------------------------
	int retval;
	static int DCount = 0;	//keeps track of the number of times its called
	ECB *e;					//pointer to event control block for semaphore
	int sr;					//save area for status register
	sr = Disable();			//disable interrupts
	char *s = NewBlock();	//get a new block of memory
	Enable(sr);				//bring interrupts back up
	sprintf(s,"DELAY%d",DCount++);		//create name for semaphore
	e = NewSemaphore(0,SEMAPHORE_MODE_TIMEOUT,s);	//create semaphore
	sr = Disable();
	DeleteBlock(s);			//return memory block
	Enable(sr);
	retval = PendSemaphore(e,mSec);	//pend for time delay period
	DeleteSemaphore(e);				//return semaphore back to pool
	return retval;					//exit
}

//--------------------------------------------------------------------------
// Initialize system
//--------------------------------------------------------------------------

void InitTask(void)
{
	InitPQ(&ActiveTasks,32,PriorityCompare);
}

